#!/usr/bin/perl
#
# $Id: run_mirs.pl 959 2015-09-17 15:16:39Z jimd $
# Use of SVN is over.  Gitlab integration - not yet!
#
# Top-level perl script for CSPP_MIRS package.
# Uses only Perl modules that are part of the v5.006 standard install.
#
# A set of user-defined functions are appended to this main script;
# they are listed in alphabetical order.
#
# Pre-requisites:
# MIRS distribution in directory $CSPP_MIRS_HOME/$CSPP_MIRS_REVNAME
# Static-linked binaries in the cspp_mirs_rNNNN/bin directory, various PCFs
# (paths & configuration files) in the cspp_mirs_rNNNN/setup directory to serve
# as templates, various SCSs (sequence control scripts) also to serve as 
# templates in the $CSPP_MIRS_REVNAME/scripts directory. Template files are
# modified to become runtime scripts are prepended 'cspp_', for example
# setup/npp_pcf.bash is used as a template to create cspp_npp_pcf.bash;
# scripts/npp_scs.bash is a template used to create cspp_npp_scs.bash.
# These two bash scripts are then used to control the end-to-end processing
# through a number of steps in the MIRS processing chain, namely ...
#
#   step_rdr2tdr                 # RDR->TDR
#   step_mergeNedt               # MERGE NEDTs
#   step_tdr2sdr                 # TDR->SDR
#   step_fm                      # FOOTPRINT MATCHING
#   step_externalDataFromRegress # USE REGRESSION ALGS TO GENERATE EXTERN DATA
#   step_fmsdr2edr               # FMSDR->EDR
#   step_prepSweClimo            # PREPARE AND GENERATE THE COLLOCATED SWE
#                                  CLIMATOLOGY FILES FOR VIPP (new for 2.0)
#   step_vipp                    # VERTICAL INTEGRATION AND POST PROCESSING
#   step_nc                      # Convert EDR & DEP into NETCDF4 format
#   step_clean                   # DISK CLEANING/PURGING
#
# ... which is the subset of MIRS processing steps used in operations.
# For step_clean I actually set it to zero and implement my own cleaning
# procedure; just remove the temporary working directory at end of job.
#
#
# START-HELP (N.B. This tag is used in printing help, do not touch!)
#
#  Options:
#  --version             show program's version number and exit
#  ?, -h, --help         show this help message and exit
#  -n, --nothing         make the PCF and SCS files, but do not execute.
#
#  Mandatory Arguments:
#    At a minimum these arguments must be specified [i.e. no default].
#
#    -i INPUT_FILES, --input_files=INPUT_FILES
#                        Fully qualified path to input files. May be directory
#                        name, or directory/filemask (use * and ?, not regexp).
#                        Filemasks MUST be "quoted" so OS does NOT expand!
#                        e.g. -i "/dataDir/{GATMO,SATMS,TATMS}*t1823486*.h5" OR
#                        -i "/dataDir/{amsua,mhs}l1b_noaa18_*_1051*l1b"      OR
#                        -i "/dataDir/*.{AMA,MHS}X.M2.D13329.S0346.E0529.B*.SV"
#                        (BUT -i /dataDir  really is the simplest way).
#    -s MISSION, --sat=MISSION
#                        The mission satellite data for MIRS algorithm.
#                        Possible values (following MIRS documentation) ...
#                        'npp', 'n18', 'n19', 'metopA', 'metopB'.
#
#  Extra Options:
#    These options may be used to customize behaviour of this program.
#
#    -w WORK_DIR, --work=WORK_DIR
#                        The directory in which all activity will occur,
#                        this is where output files will be put.
#                        [default: current directory].
#    -r RES, --resolution=RES
#                        Executes retrieval in RES resolution mode for sensors
#                        aboard 'n18', 'n19', 'metopA', and 'metopB'.  RES is
#                        'HI' or 'LO'.  When 'LO' the higher resolution sensor
#                        (MHS) is aggregated to the lower resolution sensor
#                        (AMSUA) footprint. 'HI' means AMSUA interpolated to
#                        MHS footprint.
#                        [default: 'HI'].
#    -b BLOCK, --block=BLOCK
#                        Uses bias correction file for 'npp' block BLOCK SDRs.
#                        Has no impact for 'n18', 'n19', 'metopA', 'metopB'.
#                        ADL block 2 SDRs in effect from about March 2017.
#                        [default: 2; the only other valid option is 1].
#    -x, --xfiles        Use the NOAA filenames for AMSUA/MHS L1B files
#                        [default: Direct Broadcast filenames].
#    -v, --verbose       Each occurrence increases verbosity 1 level from
#                        [default: ERROR only] -v=WARNING -vv=INFO -vvv=DEBUG
#    -c, --clean         Deletes temporary working directory under WORK_DIR
#                        at process completion.
#                        [default: --clean; --noclean is the contrary option].
#    -a, --aggregate     Aggregates npp/ATMS granules into swaths prior to
#                        processing; you ought do this (to counter cross-
#                        granule effects) unless files are already aggregated.
#                        [default: --agg; --noagg is the contrary option].
#
# END-HELP (N.B. This tag is used in printing help, do not touch!)
#
# The following option is inactive; %options{processors} defaults to 1 and is
# never interrogated. Leave here in case I ever add multicore functionality.
#
#    -p PROCESSORS, --processors=PROCESSORS
#                        Number of cpus to use for MIRS processing.
#
# NOTES-TO-SELF:
# -------------
#   $!=N to change exit code on die() to N; default is 255
#
#   --clean does not run MIRS step_clean, but achieves some of the same
#   objectives by removing the temporary directory tree. If the user
#   selects --noclean the temporary directory is NOT removed and all MIRS
#   product files are available in the temporary directory structure.
#   This option might be useful for debugging and is necessary if the
#   user wants access to product files generated prior to the output
#   netCDF files that are copied up to WORK_DIR.  It becomes fully the
#   user's responsibility dispense with any unwanted directories/files.
#
#
# This is a valid way to run a single S-NPP granule ...
#
#   ./run_mirs.pl -s npp \
#     -i "/dataDir/{GATMO,TATMS,SATMS}_npp_d20131120_t0656260_*_cspp_dev.h5"
#
# The quotes are to prevent OS expanding wildcards/tokens on the commandline.
#
# For Metop-A files with NOAA filenaming convention, something like this ...
#
#   ./run_mirs.pl -x -s metopA \
#     -i "/dataDir/NSS.{AMA,MHS}X.M2.D13329.S0346.E0529.B3684546.SV"
#
# ... will run a single overpass or orbit file.
#
# But it is a whole lot simpler to put your data into a directory and to
# process it all with something like (a DB format NOAA-18 example now) ...
#
#   ./run_mirs.pl -s n18 -i /dataDir
#


# Pragmata
# ````````
use 5.006;
use strict;
use warnings;


# Load Perl modules - standard modules only
# `````````````````````````````````````````
use File::Spec;
use File::Copy;
use Data::Dumper;
use Getopt::Long;
use File::Basename;
use Text::ParseWords;
use Encode qw(decode);
use Cwd qw(abs_path getcwd);
use List::Util qw(sum min max);
use File::Path qw(mkpath rmtree);
use Time::Local qw(timegm_nocheck);
use File::Temp qw(tempfile tempdir);
use Scalar::Util qw(looks_like_number);
use I18N::Langinfo qw(langinfo CODESET);


# Global variables are all UPPERCASE throughout
# `````````````````````````````````````````````
our $AGG_WINDOW_MINUTES = 30; # 30-minute aggregation window for npp
our $MAIN = basename($0);
defined $ENV{CSPP_MIRS_HOME} or die "$MAIN\:: CSPP_MIRS_HOME not set.\nAbort";
defined $ENV{CSPP_MIRS_REVNAME} or die "$MAIN\:: CSPP_MIRS_REVNAME not set.\nAbort";
our $HOME = File::Spec -> catfile($ENV{CSPP_MIRS_HOME}, $ENV{CSPP_MIRS_REVNAME});
(-d $HOME) or die "$MAIN\:: CSPP_MIRS subdirectory $HOME not found.\nAbort";
our $LATEST = 'working'; # root of temporary directory tree for dynamic data 


# $H5REPACK points to supplied static-compiled h5repack
# `````````````````````````````````````````````````````
# But you can point to your own h5repack if you wish. Ours is Version 1.8.11
our $H5REPACK = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'h5repack');
(-e $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not found.\nAbort";
(-x $H5REPACK) or die "$MAIN\:: h5repack $H5REPACK not executable.\nAbort";


# $H5NAGG points to supplied static-compiled nagg
# ```````````````````````````````````````````````
# But you can point to your own nagg if you wish. Ours is now 1.6.2-
our $H5NAGG = File::Spec -> catfile($ENV{'CSPP_MIRS_HOME'}, 'bin', 'nagg');
(-e $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not found.\nAbort";
(-x $H5NAGG) or die "$MAIN\:: nagg $H5NAGG not executable.\nAbort";


# The MIRS system uses just one file to specify a file group
# ``````````````````````````````````````````````````````````
# Don't mess with these unless you are confident of your actions.
# There is no ambiguity between DB and DAAC for S-NPP filenames,
# so they are hard-coded (later) in the script.
our ($AMSUA_DB, $MHS_DB) = qw(amsua mhs);
our ($AMSUA_NOAA, $MHS_NOAA) = qw(AMAX MHSX);


# Get the command line options, display help or revision if asked
# ```````````````````````````````````````````````````````````````
my $revision = $ENV{CSPP_MIRS_REVDATE};
my %options = get_commandline($revision);


# These are used for fileglobbing when INPUTFILES is a directory name
# ```````````````````````````````````````````````````````````````````
# Change to suit your site; for more control edit function validate_options(),
# or provide "<dirname>/<my_filemask>" on the command line as INPUTFILES; the
# "quotes" are important to prevent the OS expanding wildcard tokens.
my $group = "{$AMSUA_DB,$MHS_DB}";
my %filemask = (npp => '{GATMO,SATMS,TATMS}*.h5',
                n18 => $group.'l1b_noaa18*.l1b',
                n19 => $group.'l1b_noaa19*.l1b',
                metopA => $group.'l1b_M02*.l1b',
                metopB => $group.'l1b_M01*.l1b');
($options{xfiles}) and do {
  $group = "{$AMSUA_NOAA,$MHS_NOAA}";
  %filemask = (npp => '{GATMO,SATMS,TATMS}*.h5',
               n18 => '*.'.$group.'.NN*',
               n19 => '*.'.$group.'.NP*',
               metopA => '*.'.$group.'.M2*',
               metopB => '*.'.$group.'.M1*'); };
($options{verbose} == 3) and pretty_dump(%filemask);

# Parse and validate command line %options & make temporary working directory
# ```````````````````````````````````````````````````````````````````````````
%options = validate_options($filemask{$options{sat}}, %options);
my $tmpdir = tempdir('TMP_XXXXXX', DIR => $options{work}, CLEANUP => $options{clean});
my $pwd = getcwd();
chdir($tmpdir) or
  die "$MAIN\:: Cannot chdir to temporary working directory $tmpdir\nAbort";
($options{verbose} >= 2) and
  print "chdir to working directory $tmpdir\n";


# Make PCF and SCS scripts for this runtime environment
# `````````````````````````````````````````````````````
my $pcf = make_PCF($tmpdir, %options);
my $scs = make_SCS($tmpdir, $pcf, %options);
($options{verbose} >= 2) and do {
  print "PCF file is:  $pcf\n";
  print "SCS file is:  $scs\n"; };


# Preprocess and stage data for MIRS system compatibility
# ``````````````````````````````````````````````````````
%options = cspp_preprocess(%options);
($options{verbose} == 3) and pretty_dump(%options);


# Run the job (worry later about specifying nCPUs)
# ````````````````````````````````````````````````
exec_SCS($scs, %options); # internal abort on error


# Leave a spare directory tree - useful for debugging
# ```````````````````````````````````````````````````
($options{clean}) or make_tmpdir_tree($options{sat});


# All done - chdir so that temporary working directory can be disappeared
# ```````````````````````````````````````````````````````````````````````
chdir($pwd) or die "$MAIN\:: Cannot chdir back to $pwd\nAbort";
exit(0);



###############################################################################
#                                                                             #
#         User-defined functions - they are listed alphabetically             #
#                                                                             #
#   NOTE: My editor can collapse code between matching {} braces, this        #
#   explains the strange formatting - I can collapse each function to see     #
#   just the function name and an example of how to call it; and expand       #
#   to see the more detailed documentation and to edit.      JED Nov 2013     #
#                                                                             #
###############################################################################

# -------------------
  sub cspp_preprocess
# ===================
# %options = cspp_preprocess(%options);
  {
    # On input, %options comprises some command line parameters plus other
    # processing options, including defaults for unspecified command line values
    # and extra stuff defined in main to guide processing. On output, %options
    # needs to include a field $options{process_files} that points to files
    # that satisfy MIRS system requirements in terms of both filename and
    # internal format.
    # The old nagg command was :
    # my $cmd = "$H5NAGG -S -t TATMS,SATMS -n 150 -d . $granules/*_npp_*.h5";
    # This sometimes caused pre-processsing aggregation into two files per
    # swath instead of one because of an enforced break at some time or partial
    # orbit boundaries.  The newer versions of nagg have a --onefile option to
    # deal with this occurence.
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @process_files = ();
    ($options{sat} eq 'npp') and do {
      if ($options{aggregate}) {
        my %swaths = npp_granules_to_swaths(@{$options{input_files}});
        my $iswath = 0;
        foreach my $key (sort keys %swaths) {
          my $granules = sprintf("granules_%03d", ++$iswath);
          mkpath($granules) or die "$name\: WORK_DIR $granules cannot be made.\nAbort";
          foreach (@{$swaths{$key}}) {
            my $source = File::Spec -> catfile($options{input_dir}, $_);
            my $dest = File::Spec -> catfile($granules, $_);
            my $cmd = "$H5REPACK -n -f NONE $source $dest";
            system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
          }
          my $cmd = "$H5NAGG -S -t TATMS,SATMS --onefile -d . $granules/*_npp_*.h5";
          system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
        }
      } else {
        foreach (@{$options{input_files}}) {
          my $source = File::Spec -> catfile($options{input_dir}, $_);
          my $cmd = "$H5REPACK -n -f NONE $source $_";
          system($cmd) == 0 or die "$name\: Could not run command $cmd\nAbort";
        }
      }
      my @gatmo = sort(glob('GATMO*'));
      my @satms = sort(glob('SATMS*'));
      my @tatms = sort(glob('TATMS*'));
      ($#tatms == $#gatmo) or die "$name\: Number of TATMS != GATMO\nAbort";
      ($#satms == $#tatms) or die "$name\: Number of SATMS != TATMS\nAbort";
      @process_files = @tatms; };
    (grep {$options{sat} eq $_} qw/n18 n19 metopA metopB/) and do {
      foreach (@{$options{input_files}}) {
        my $pf = undef;
        my $df = File::Spec -> catfile($options{input_dir}, $_);
        (index($_, $AMSUA_DB) != -1 or index($_, $AMSUA_NOAA) != -1) and do {
          $pf = make_L1B_bigEndian('amsua', $df); };
        (index($_, $MHS_DB) != -1 or index($_, $MHS_NOAA) != -1) and do {
          $pf = make_L1B_bigEndian('mhs', $df); };
        defined($pf) or die "$name\: Could not endian convert $_\nAbort";
        $options{site_id} = substr($pf, 0, 3);
        my $mf = 'NSS'.substr($pf, 3);
        $mf =~ s/\.MB\./\.M1\./; $mf =~ s/\.MA\./\.M2\./;
         ($mf eq $pf) or do { # avoid link-to-self (which will fail!)
          symlink($pf, $mf) or die "$name\: Cannot link $pf\nAbort"; };
        (index($pf, $AMSUA_NOAA) != -1) and push(@process_files, $mf);
      } 
      my @amsua = sort(glob("NSS.$AMSUA_NOAA.*"));
      my @mhs = sort(glob("NSS.$MHS_NOAA.*"));
      my $n = $#amsua;
      ($n == $#mhs) or die "$name\: Number of $AMSUA_NOAA != $MHS_NOAA\nAbort";
      for (my $i=0; $i<=$n; $i++) {
        my ($a1, $Sa, $a2, $Ea, $a3) = unpack('a20a4a2a4', $amsua[$i]);
        (looks_like_number($Sa) and looks_like_number($Ea)) or
          die  "$name\: $options{sat} $amsua[$i] unrecognized\nAbort";
        my ($m1, $Sm, $m2, $Em, $m3) = unpack('a20a4a2a4a12', $mhs[$i]);
        (looks_like_number($Sm) and looks_like_number($Em)) or
          die  "$name\: $options{sat} $mhs[$i] unrecognized\nAbort";
        ((abs($Sa-$Sm) + abs($Ea-$Em)) > 0) and do {
          ($options{verbose} > 0) and 
            warn "$name\: WARNING! $AMSUA_NOAA:$MHS_NOAA mismatch. Forcing\n";
          my $forced = pack('a20a4a2a4a12', $m1, $Sa, $m2, $Ea, $m3);
          move("$mhs[$i]", "$forced"); };
      } };
    @{$options{process_files}} = @process_files;
    return %options;
  }

# ------------
  sub exec_SCS
# ============
# exec_SCS($SCS, %options);
  {
    # $SCS is the runtime SCS file created by make_SCS()
    # %options contains things I need like the SDR/L1B files to
    # process plus verbosity and do-nothing choices. I do not need
    # the name of the PCF because it is already named in the SCS.
    my $SCS = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    ($options{nothing} and ($options{verbose} > 0)) and
      print "$name\: Do not execute mode --nothing\n";
    $name =~ s/main/$MAIN/g;
    my $globber = "NSS.$AMSUA_NOAA*";
    ($options{sat} eq 'npp') and $globber = "TATMS*";
    my $nglob = 0;
    foreach (glob($globber)) {
      my $parked = sprintf("swath_%03d", ++$nglob);
      my $outdir = make_tmpdir_tree($options{sat});
      my $cmd = sprintf("%s %s", $SCS, $_);
      ($options{verbose} > 0) and print "$cmd\n";
      ($options{nothing}) or do {
        system($cmd)==0 or die "$name\: Cannot complete $cmd\nAbort"; };
      my $ncglobber = File::Spec -> catfile($outdir, '*.nc');
      my $loglobber = '*logFile*.dat';
      foreach ((glob($ncglobber), glob($loglobber))) {
        copy($_, $options{work}) or die "$name\: Copy failed: $!\nAbort";
      }
      move("$LATEST", "$parked");
    }
    return;
  }

# -------------------
  sub get_commandline
# ===================
# %options = get_commandline($revision);
  {
    # If things go bad in here, jump to usage() which terminates with exit(9).
    # Note that I had to hack in the Python-like -vv & -vvv support.
    # This routine assumes standard module Getopt::Long has been loaded.
    # abs_path() is dangerous for the purpose of setting the work (and,
    # equivalently, the output) directory - see this post:
    # http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=257568
    my $rev = shift;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $codeset = langinfo(CODESET);
    @ARGV = map{decode $codeset, $_} @ARGV;
    foreach (@ARGV) { s/\?/-\?/g; }
    my %options = ();
    GetOptions(\%options,
      'version|V',
      'help|?',
      'nothing|n',
      'input_files|i=s',
      'sat|s=s',
      'work|w=s',
      'resolution|r=s',
      'block|b=i',
      'processors|p=i',
      'verbose|v+', 'vv', 'vvv',
      'xfiles|x',
      'aggregate|a!',
      'clean|c!') or usage();
    defined($options{version}) and die "$rev\n";
    defined($options{help}) and usage();
    $options{nothing} = defined($options{nothing}) ? $options{nothing} : 0;
    $options{xfiles} = defined($options{xfiles}) ? $options{xfiles} : 0;
    defined($options{input_files}) or usage();
    defined($options{sat}) or usage();
    (grep {$_ eq $options{sat}} qw/npp n18 n19 metopA metopB/) or usage();
    my $cwd = getcwd();
    (defined($options{work})) or $options{work} = $cwd;
    (substr($options{work}, 0, 1) eq '/') or do {
      $options{work} = File::Spec -> catfile($cwd, $options{work}); };
    (defined($options{resolution})) or $options{resolution} = 'HI';
    (defined($options{block})) or $options{block} = 2;
    $options{processors} = defined($options{processors})
      ? $options{processors} : 1; # not currently used
    $options{verbose} = defined($options{verbose}) ? $options{verbose} : 0;
    $options{vv} = defined($options{vv}) ? $options{vv} : 0;
    $options{vvv} = defined($options{vvv}) ? $options{vvv} : 0;
    $options{verbose} = $options{verbose} + 2 * $options{vv};
    $options{verbose} = $options{verbose} + 3 * $options{vvv};
    $options{verbose} = min($options{verbose}, 3);
    delete @options{'vv', 'vvv'};
    $options{clean} = defined($options{clean}) ? $options{clean} : 1;
    $options{aggregate} = defined($options{aggregate}) ? $options{aggregate} : 1;
    return %options;
  }

# -----------------------------------------------------------
  sub is_integer { defined $_[0] and $_[0] =~ /^[+-]?\d+$/; }
# ===========================================================

# -------------------------
  sub is_plausible_datetime
# =========================
# $status = is_plausible_date($YYYY, $MM, $DD, $hh, $mm);
  {
    # NOTE: Year < 2000 is considered implausible in this function.
    my ($YYYY, $MM, $DD, $hh, $mm) = (@_);
    my $implausible_year = 2000;
    is_integer($YYYY) or return 0;
    $YYYY < $implausible_year and return 0;
    is_integer($MM) or return 0;
    $MM >= 1 and $MM <= 12 or return 0;
    is_integer($DD) or return 0;
    $DD >= 1 and $DD <= 31 or return 0;
    is_integer($hh) or return 0;
    $hh >= 0 and $hh <= 23 or return 0;
    is_integer($mm) or return 0;
    $mm >= 0 and $mm <= 59 or return 0;
    return 1;
  }

# ----------------------
  sub make_L1B_bigEndian
# ======================
# $outfile = make_L1B_bigEndian($instrument, $infile);
  {
    # $instrument is one of 'amsua' or 'mhs'
    # $infile is the name of the L1B file to test/convert endian
    # $outfile is the big endian output filename created
    #
    # $outfile is extracted from the header of $infile, bytes 23-64 inclusive
    # (42 bytes). The reason for creating $outfile in WORK_DIR/tempdir is so
    # that we don't clobber data files in WORK_DIR if the user happens to
    # specify WORK_DIR as the place where the original L1B files are stored.
    # For files named like "amsual1b_noaa18_20131014_2145_43295.l1b" no problem,
    # but for files named like "NSS.AMAX.NN.D06032.S0122.E0317.B0361819.WI",
    # problem since this is the same as the filename stored in its header!
    #
    # NOTE: We coalesce repeated character variables to strings in the $Lhdr and
    # $Lrec buffer specification (because we are not actually applying any logic
    # to characters), but only after 'a42a8' in $Lhdr, since the a42 field is to
    # be used to define $outfile.
    #
    my ($instrument, $infile) = (shift, shift);
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my ($Lhdr, $Bhdr, $Lrec, $Brec, $count, $buf) = undef;
    my ($reclen, $year, $iyear, $doy, $idoy, $iout) = undef;
    ($instrument eq 'amsua') and do {
      ($reclen, $iout, $iyear, $idoy) = (2560, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv3v5Vv2V2v2Vv14Va16V4v106V115a8v8V363v6Vv2V46';
      $Lrec = 'v4Vv2V4v16V95v6V2v94V61v620Vv2V2a28Vv156Vv2V2a32v6V'; };
    ($instrument eq 'mhs') and do {
      ($reclen, $iout, $iyear, $idoy) = (3072, 11, 22, 23);
      $Lhdr = 'a3av9a42a8v4Vv2V2v2Vv8Vv2Vv15a16v51V53a8v8V92v436Vv2V6v642';
      $Lrec = 'v4Vv2VvaaV2v14V30v8V2v274V180v544V2v48V2a78v9V5a52v12Vv102'; };
    defined($reclen) or die "$name\: Instrument $instrument not supported\nAbort";
    $Bhdr = $Lhdr; $Bhdr =~ s/v/n/g; $Bhdr =~ s/V/N/g;
    $Brec = $Lrec; $Brec =~ s/v/n/g; $Brec =~ s/V/N/g;
    open my $fh, "<$infile"; binmode $fh;
    $count = read($fh, $buf, $reclen);
    ($count == $reclen) or die "$name\: Header count != reclen $infile\nAbort";
    my @data = unpack($Lhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Lendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    @data = unpack($Bhdr, $buf);
    ($year, $doy) = ($data[$iyear], $data[$idoy]);
    my $Bendian = ($year > 1960 and $year < 2099 and $doy > 0 and $doy < 367);
    ($Lendian xor $Bendian) or die "$name\: $infile endian indeterminate\nAbort";
    my $outfile = $data[$iout];
    $outfile =~ s/^\s+//; $outfile =~ s/\s+$//; $outfile =~ s/ //g;
    (length($outfile) != 42) and die "$name\: L1B filename $outfile funky\nAbort";
    open my $FH, ">$outfile"; binmode $FH;
    ($Lendian) and $buf = pack($Bhdr, unpack($Lhdr, $buf));
    print $FH $buf;
    while (!eof($fh)) {
      $count = read($fh, $buf, $reclen);
      ($count == $reclen) or die "$name\: Record count != reclen $infile\nAbort";
      ($Lendian) and $buf = pack($Brec, unpack($Lrec, $buf));
      print $FH $buf;
    }
    close $fh; close $FH;
    return $outfile;
  }

# ------------
  sub make_PCF
# ============
# $pcf = make_PCF($tmpdir, %options);
  {
    # $tmpdir is fully-qualified pathname of temporary working directory.
    # %options are the command line options, parsed and validated.
    # $pcf is fully-qualified PCF filepath for this run.
    # $LATEST is name of the working directory within $tmpdir for the
    #   swath or granule currently being processed - really it is fixed
    #   in MAIN as 'working'.
    # $latest is the fully-qualified version of $LATEST
    #
    # The problem with $latest is that for some processes (at least for
    # RUNNING TDR->SDR(atms) STEP), the fully qualified filenames that are
    # prefixed by $latest get truncated to 200 characters in the MIRS system.
    # This can be a problem, not least because the filenames themselves can
    # be over 80 characters long. Since we have already chdir'd into $tmpdir,
    # we can use relative paths.
    #
    # Changes for CSPP_MIRS_2_0 arise from new functionality AND because I
    # have decided to NOT pre-edit any of the supplied PCF files in the
    # delivery.  The $CSPP_MIRS_REVNAME environment is now the dirname that
    # the DAP unpacks into (in CSPP_MIRS_1_0 I prefixed with cspp_) and I just
    # pick an existant PCF (i.e. as delivered) to use as a template.
    #
    my $tmpdir = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $tmp = File::Spec -> catfile($HOME, 'setup', $options{sat}.'_pcf_benchmark_hr.bash');
    -e $tmp or die "$MAIN\:: PCF template $tmp does not exist.\nAbort";
    my $pcf = File::Spec -> catfile($tmpdir, 'cspp_'.$options{sat}.'_pcf.bash');
    open(my $fhI, "<$tmp") or die "$name\: Cannot open $tmp for read.\nAbort";
    open(my $fhO, ">$pcf") or die "$name\: Cannot open $pcf for write.\nAbort";
    while (<$fhI>) { chomp;
      # text swaps from CSPP_MIRS_1_0
      (/date=/) and $_='date=dummy';
      (/rootPath=/) and $_='rootPath='.$HOME;
      (/rdrSensor1Path=/) and $_='rdrSensor1Path=./';
      (/rdrSensor2Path=/) and $_='rdrSensor2Path=./';
      (/rdrOrbitPath=/) and $_='rdrOrbitPath=./';
      (/controlDataPath=/) and $_='controlDataPath='.$LATEST.'/ControlData';
      (/inputDataPath=/) and $_='inputDataPath='.$LATEST.'/InputsData';
      (/testbedDataPath=/) and $_='testbedDataPath='.$LATEST.'/TestbedData';
      (/step_clean=/) and $_='step_clean=0'; # MIRS cleaning not good for CSPP.
      (/useCPU=/) and $_='useCPU=0'; # Manage cores at script level - if ever.
      # text swaps new for CSPP_MIRS_2_0
      (/logPath=/) and $_='logPath=./';
      (/step_choppRadFiles=/) and $_='step_choppRadFiles=0';
      (/step_mergeEdr=/) and $_='step_mergeEdr=0';
      (/retrOnOrbitOrSubOrbit=/) and $_='retrOnOrbitOrSubOrbit=0';
      (/outFMAccuracy=/) and $_='outFMAccuracy=0';
      (/fmType=/) and do { ($options{resolution} eq 'LO') and $_='fmType=0'; };
      # text swaps new for CSPP_MIRS_2_0 at patch level 2.0.2
      ($options{sat} eq 'npp') and do {
        ($options{block} == 1) and do {
          (/biasFileToUse=/) and $_='biasFileToUse=${biasPath}/biasCorrecBlock1_${satId}.dat';
        };
        ($options{block} == 2) and do {
          (/biasFileToUse=/) and $_='biasFileToUse=${biasPath}/biasCorrecBlock2_${satId}.dat';
        };
      };
      print $fhO "$_\n";
    }
    close($fhI) or die "$name\: Cannot close file $tmp.\nAbort";
    close($fhO) or die "$name\: Cannot close file $pcf.\nAbort";
    return $pcf;
  }

# ------------
  sub make_SCS
# ============
# $scs = make_SCS($tmpdir, $pcf, %options);
  {
    # $tmpdir is fully-qualified pathname of temporary working directory.
    # $pcf is fully-qualified PCF filepath for this run.
    # %options are the command line options, parsed and validated.
    # $scs is fully-qualified SCS filepath for this run.
    my ($tmpdir, $pcf) = (shift, shift);
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $setup = File::Spec -> catfile($HOME, 'setup');
    my $scripts = File::Spec -> catfile($HOME, 'scripts');
    foreach ($setup, $scripts) {
      -d or die "$name\:: MIRS subdirectory $_ not found.\nAbort";
    }
    my $tmp = File::Spec -> catfile($scripts, $options{sat}.'_scs_benchmark_hr.bash');
    -e $tmp or die "$MAIN\:: SCS template $tmp does not exist.\nAbort";
    my $scs = File::Spec -> catfile($tmpdir, 'cspp_'.$options{sat}.'_scs.bash');
    open(my $fhI, "<$tmp") or die "$name\: Cannot open $tmp for read.\nAbort";
    open(my $fhO, ">$scs") or die "$name\: Cannot open $scs for write.\nAbort";
    while (<$fhI>) { chomp;
      (/benchmark_hr.bash/) and $_ ='. '.$pcf; # i.e. source the PCF
      (/..\/setup/) and s/..\/setup/$setup/g;
      (/..\/scripts/) and s/..\/scripts/$scripts/g;
      print $fhO "$_\n";
    }
    close($fhI) or die "$name\: Cannot close file $scs\nAbort";
    close($fhO) or die "$name\: Cannot close file $tmp\nAbort";
    chmod (0755, $scs) or die "$name\: Cannot chmod $scs.\nAbort";
    return $scs;
  }

# --------------------
  sub make_tmpdir_tree
# ====================
# $outdir = make_tmpdir_tree($sat);
  {
    # Make ControlData, InputsData and TestbedData directory trees under
    #   cwd - needed for processing $sat data with MIRS, and without
    #   clashing in the distribution directories (which will otherwise
    #   happen for two+ processes running concurrently).
    # $outdir is the directory that will contain output netCDF files.
    # Aborts within on fail.
    # It looks like the delivery from NOAA/STAR will make directories anyway,
    # but I want to be clear on the directories actually needed. The delivery
    # may make extra directories that appear to be unused.
    my $sat = shift;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my @dirs = ();
    foreach (qw/ControlData InputsData TestbedData/) {
      /ControlData/ and push(@dirs, File::Spec -> catfile($LATEST, $_));
      /InputsData/ and push(@dirs, File::Spec -> catfile($LATEST, $_));
      /TestbedData/ and do {
        my $d1 = $_;
        foreach (qw/DynamicData nedt Outputs/) {
          /DynamicData/ and do {
            my $d2 = $_;
            foreach (qw/fmsdr regress_retr sdr swe_climo tdr/) {
              my $d3 = $_;
              push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            } };
          /nedt/ and do {
            my $d2 = $_;
            my $d3 = undef;
            ($sat eq 'npp') and $d3 = 'npp_atms';
            ($sat eq 'n18') and $d3 = 'n18_amsua';
            ($sat eq 'n19') and $d3 = 'n19_amsua';
            ($sat eq 'metopA') and $d3 = 'metopA_amsua';
            ($sat eq 'metopB') and $d3 = 'metopB_amsua';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            ($sat eq 'npp') and next;
            ($sat eq 'n18') and $d3 = 'n18_mhs';
            ($sat eq 'n19') and $d3 = 'n19_mhs';
            ($sat eq 'metopA') and $d3 = 'metopA_mhs';
            ($sat eq 'metopB') and $d3 = 'metopB_mhs';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3));
            ($sat eq 'npp') and next;
            ($sat eq 'n18') and $d3 = 'n18_amsua_mhs';
            ($sat eq 'n19') and $d3 = 'n19_amsua_mhs';
            ($sat eq 'metopA') and $d3 = 'metopA_amsua_mhs';
            ($sat eq 'metopB') and $d3 = 'metopB_amsua_mhs';
            push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3)); };
          /Outputs/ and do {
            my $d2 = $_;
            foreach (qw/dep edr nc/) {
              my $d3 = $_;
              my $d4 = undef;
              ($sat eq 'npp') and $d4 = 'npp_atms';
              ($sat eq 'n18') and $d4 = 'n18_amsua_mhs';
              ($sat eq 'n19') and $d4 = 'n19_amsua_mhs';
              ($sat eq 'metopA') and $d4 = 'metopA_amsua_mhs';
              ($sat eq 'metopB') and $d4 = 'metopB_amsua_mhs';
              push(@dirs, File::Spec -> catfile($LATEST, $d1, $d2, $d3, $d4));
            } };
        } };
    }
    foreach (@dirs) { mkpath($_) or die "$name\: Cannot make dir $_\nAbort"; }
    return $dirs[$#dirs]; # last dir is netCDF output dir
  }


# --------------------------
  sub npp_granules_to_swaths
# ==========================
# %swaths = npp_granules_to_swaths(@files);
  {
    # Break list of files into groups that comprise contiguos overpass
    # swaths ready for the hdf4 aggregator (nagg) to concatenate.
    my @files = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $time_window = 60 * $AGG_WINDOW_MINUTES;
    my %swaths = ();
    my ($YYYY, $MM, $DD, $hh, $mm, $found) = undef;
    foreach (@files) {
      $YYYY = substr($_, 11, 4);
      $MM = substr($_, 15, 2);
      $DD = substr($_, 17, 2);
      $hh = substr($_, 21, 2);
      $mm = substr($_, 23, 2);
      is_plausible_datetime($YYYY, $MM, $DD, $hh, $mm) or
        die "$name\: Date from npp file $_ implausible\nAbort";
      my $t = timegm_nocheck(0, $mm, $hh, $DD, $MM-1, $YYYY);
      $found = 0;
      foreach my $key (sort keys %swaths) {
        (abs($key - $t) < $time_window) and do {
          push(@{$swaths{$key}}, $_);
          $found = 1; last; };
      }
      ($found) or @{$swaths{$t}} = ($_);
    }
    return %swaths;
  }

# ---------------
  sub pretty_dump
# ===============
# pretty_dump(%hash);
  {
    # Do a "pretty print" of %hash by effectively grepping Dumper().
    # Assumes standard module Data::Dumper has been loaded.
    my (%hash) = @_;
    foreach (split('\n', Dumper(\%hash))) {
      chomp;
      ((/VAR1/) or (/\}\;/)) and next;
      $_ =~ s/  //g; # shorten up the whitespace
      print "$_\n";
    }
  }

# ---------
  sub usage
# =========
# usage();
  {
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    open(my $fh, "<$0");
    while (<$fh>) { chomp;
      if ( /START-HELP/ ) {
        while (<$fh>) { chomp;
          ( /END-HELP/ ) and do { close($fh); exit(9); };
          my $help = (split(/\#/, $_))[1]; # remove leading hash
          $help and print "$help\n" or print "\n";
        }
      }
    }
    die "$name\: Unprecedented error!\nAbort";
  }

# --------------------
  sub validate_options
# ====================
# %options = validate_options($filemask, %options);
  {
    # $filemask is for globbing the nominated satellite data.
    # %options are the command line options to "validate".
    # Fatal errors in here terminate with die without showing usage() message.
    # The most important task is to convert $options{input_files} to a scalar
    # $options{input_dir} and filename list put back in $options{input_files}.
    my $filemask = shift;
    my (%options) = @_;
    my $name = (caller(0))[3];
    $name =~ s/main/$MAIN/g;
    my $globber = $options{input_files};
    (-d $options{input_files}) and do {
      my $base = $options{input_files};
      $globber = File::Spec -> catfile($base, $filemask); };
    my @input_files = sort(glob($globber));
    ($#input_files < 0) and die "$name\: INPUTFILES not found.\nAbort";
    $options{input_dir} = dirname(abs_path($input_files[0]));
    $options{input_files} = ();
    foreach (@input_files) { push(@{$options{input_files}}, basename($_)); }
    (-d $options{work}) or do {
      ($options{verbose} >=2) and print "Making WORK_DIR $options{work}\n";
      mkpath($options{work}) or
        die "$name\: WORK_DIR $options{work} cannot be made.\nAbort"; };
    (-w $options{work}) or
      die "$name\: WORK_DIR $options{work} not writable.\nAbort";
    (grep {$options{resolution} eq $_} qw/h H hi HI hI Hi/) and do {
      $options{resolution} = 'HI'; };
    (grep {$options{resolution} eq $_} qw/l L lo LO lO Lo/) and do {
      $options{resolution} = 'LO'; };
    (grep {$options{resolution} eq $_} qw/HI LO/) or
      die "$name\: RES $options{resolution} invalid resolution.\nAbort";
    (($options{sat} eq 'npp') and ($options{resolution} eq 'LO')) and
      die "$name\: RES $options{resolution} invalid for 'npp'.\nAbort";
    (($options{block} >= 1) and ($options{block} <= 2)) or
      die "$name\: BLOCK $options{block} invalid ADL block.\nAbort";
    return %options;
  }
